#include <avr/pgmspace.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/io.h>

#include "game.h"
#include "tetrominoes.h"

// Vectorised board matrix
static uint8_t game_board[BOARD_ROWS * BOARD_COLS];
static struct tetro curr_tetro;

static uint8_t level;
static uint32_t score;
static uint32_t high_score;
static uint8_t lines_until_next_level;

// Time until the falling tetromino should go down one position.
// Time differs from level to level, for more info go to:
// https://tetris.fandom.com/wiki/Tetris_(NES,_Nintendo)
static uint16_t OCR_values_per_level[] = {
	12499, 11197, 9895, 8592, 7290, 5988, 4686, 3383, 2081, 1560, // 00 - 09
	1300, // 10 - 12
	1039, // 13 - 15
	779,  // 16 - 18
	518,  // 19 - 28
	258   // 29+
};

static enum game_state state;

// Variables to keep track if a button press or hold event was treated or not
// 1 - event not treated
// 0 - even treated
static volatile uint8_t up_button_pressed, up_button_hold;
static volatile uint8_t down_button_pressed, down_button_hold;
static volatile uint8_t left_button_pressed, left_button_hold;
static volatile uint8_t right_button_pressed, right_button_hold;
static volatile uint8_t ccw_button_pressed, ccw_button_hold;
static volatile uint8_t cw_button_pressed, cw_button_hold;
// Keep track if current falling tetromino should be moved down one row
static volatile uint8_t down_timer_expired;


ISR(TIMER1_COMPA_vect)
{
	down_timer_expired = 1;
}

static void timer1_init(void)
{
	TCNT1 = 0;
	OCR1A = OCR_values_per_level[0];
	TIMSK1 = (1 << OCIE1A);
	TCCR1B = (5 << CS10) | (1 << WGM12);
}

/**
 * Create a new random tetromino at the top of the board.
 */
static void spawn_new_tetro(void)
{
	enum tetro_type type = rand() % TETRO_NUM;

	memset(&curr_tetro, 0, sizeof(struct tetro));

	switch (type) {
	case T_I:
		curr_tetro = tetro_i;
		break;
	case T_J:
		curr_tetro = tetro_j;
		break;
	case T_L:
		curr_tetro = tetro_l;
		break;
	case T_O:
		curr_tetro = tetro_o;
		break;
	case T_S:
		curr_tetro = tetro_s;
		break;
	case T_T:
		curr_tetro = tetro_t;
		break;
	case T_Z:
		curr_tetro = tetro_z;
		break;
	default:
		curr_tetro = tetro_none;
	}

	curr_tetro.rot = rand() % curr_tetro.max_rot;
	curr_tetro.row = BOARD_ROWS - 1;
	curr_tetro.col = BOARD_COLS / 2 - curr_tetro.size / 2;
}

/**
 * Generate a random number by reading data from an unconnected pin.
 */
static int generate_seed(void)
{
	ADMUX = (1 << REFS0);
	ADCSRA = (1 << ADEN) | (5 << ADPS0);
	ADMUX = (ADMUX & ~(0x1f << MUX0)) | 1;

	ADCSRA |= (1 << ADSC);
	while (ADCSRA & (1 << ADSC));

	return ADC;
}

/**
 * Test if moving the current tetromino to a new position will cause the
 * tetromino to overlap with other minoes or with the walls.
 * @param row row coordinate of position to test
 * @param col column coordinate of position to test
 * @param rot rotation of the tetromino
 * @return 1 if collision detected, 0 otherwise
 */
static uint8_t test_collision(int8_t row, int8_t col, uint8_t rot)
{
	int r, c;

	for (r = 0; r < curr_tetro.size; ++r) {
		for (c = 0; c < curr_tetro.size; ++c) {
			if (tetro_get_data(&curr_tetro, r, c, rot)) {
				int mino_row = row + r;
				int mino_col = col + c;
				uint8_t cond1 = mino_row < 0 || mino_col < 0 ||
					mino_col >= BOARD_COLS;
				uint8_t cond2 = mino_row < BOARD_ROWS &&
					game_board[mino_row * BOARD_COLS + mino_col];

				if (cond1 || cond2) {
					return 1;
				}
			}
		}
	}

	return 0;
}

/*****************************************************************************/
/*                       Move / Rotate / Land Tetromino                      */
/*****************************************************************************/
static uint8_t tetro_can_move_down(void)
{
	return !test_collision(curr_tetro.row - 1, curr_tetro.col, curr_tetro.rot);
}

static uint8_t tetro_can_move_left(void)
{
	return !test_collision(curr_tetro.row, curr_tetro.col - 1, curr_tetro.rot);
}

static uint8_t tetro_can_move_right(void)
{
	return !test_collision(curr_tetro.row, curr_tetro.col + 1, curr_tetro.rot);
}

static uint8_t tetro_can_rotate_ccw(void)
{
	return !test_collision(curr_tetro.row, curr_tetro.col,
		(curr_tetro.rot + 1) % curr_tetro.max_rot);
}

static uint8_t tetro_can_rotate_cw(void)
{
	return !test_collision(curr_tetro.row, curr_tetro.col,
		(curr_tetro.rot - 1 + curr_tetro.max_rot) % curr_tetro.max_rot);
}

static uint8_t tetro_move_down(void)
{
	if (tetro_can_move_down()) {
		--curr_tetro.row;
		return 1;
	}
	return 0;
}

static uint8_t tetro_move_left(void)
{
	if (tetro_can_move_left()) {
		--curr_tetro.col;
		return 1;
	}
	return 0;
}

static uint8_t tetro_move_right(void)
{
	if (tetro_can_move_right()) {
		++curr_tetro.col;
		return 1;
	}
	return 0;
}

static uint8_t tetro_rotate_ccw(void)
{
	if (tetro_can_rotate_ccw()) {
		curr_tetro.rot = (curr_tetro.rot + 1) % curr_tetro.max_rot;
		return 1;
	}
	return 0;
}

static uint8_t tetro_rotate_cw(void)
{
	if (tetro_can_rotate_cw()) {
		curr_tetro.rot = (curr_tetro.rot - 1 + curr_tetro.max_rot)
			% curr_tetro.max_rot;
		return 1;
	}
	return 0;
}

static void land_tetro(void)
{
	int r, c;

	for (r = 0; r < curr_tetro.size; ++r) {
		for (c = 0; c < curr_tetro.size; ++c) {
			if (tetro_get_data(&curr_tetro, r, c, curr_tetro.rot)) {
				int mino_row = curr_tetro.row + r;
				int mino_col = curr_tetro.col + c;

				game_board[mino_row * BOARD_COLS + mino_col] = 1;
			}
		}
	}
}

static int clear_filled_rows(void)
{
	int r, r_up, c, mino_count, removed_rows = 0;

	for (r = 0; r < BOARD_ROWS; ++r) {
		// Count all minoes from the current row
		mino_count = 0;
		for (c = 0; c < BOARD_COLS; ++c) {
			if (game_board[r * BOARD_COLS + c]) {
				++mino_count;
			}
		}

		if (mino_count == BOARD_COLS) {
			// Move all minoes which are above the deleted row one
			// row downwards
			++removed_rows;
			for (r_up = r; r_up < BOARD_ROWS - 1; ++r_up) {
				for (c = 0; c < BOARD_COLS; ++c) {
					game_board[r_up * BOARD_COLS + c] =
						game_board[(r_up + 1) * BOARD_COLS + c];
				}
			}
			// Clear the last row
			memset(&game_board[(BOARD_ROWS - 1) * BOARD_COLS], 0, 
				BOARD_COLS * sizeof(uint8_t));
			--r;
		}
	}
	return removed_rows;
}

/*****************************************************************************/
/*                           Leveling up and score                           */
/*****************************************************************************/
// For more information about leveling up and score go to:
// https://tetris.fandom.com/wiki/Tetris_(NES,_Nintendo)
static uint16_t get_OCR_value_level(uint8_t level)
{
	if (level >= 0 && level <= 9) {
		return OCR_values_per_level[level];
	}
	if (level >= 10 && level <= 12) {
		return OCR_values_per_level[10];
	}
	if (level >= 13 && level <= 15) {
		return OCR_values_per_level[11];
	}
	if (level >= 16 && level <= 18) {
		return OCR_values_per_level[12];
	}
	if (level >= 19 && level <= 28) {
		return OCR_values_per_level[13];
	}
	return OCR_values_per_level[14];
}

static uint8_t get_lines_until_level(uint8_t level)
{
	if (level >= 0 && level <= 9) {
		return 10 * (level + 1);
	}
	if (level >= 10 && level <= 15) {
		return 100;
	}
	if (level >= 16 && level <= 24) {
		return (level - 5) * 10;
	}
	return 200;
}

static void score_add_drop_points(void)
{
	score++;
}

static void score_add_clear_rows(uint8_t rows)
{
	switch (rows) {
	case 1:
		score += 40 * (level + 1);
		break;
	case 2:
		score += 100 * (level + 1);
		break;
	case 3:
		score += 300 * (level + 1);
		break;
	case 4:
		score += 1200 * (level + 1);
	}
}

static void level_up_if_needed(int clr_rows)
{
	if (clr_rows >= lines_until_next_level) {
		// Avoid underflow by adding next level rows
		uint8_t next_level_lines = get_lines_until_level(level + 1);

		lines_until_next_level += next_level_lines - clr_rows;
		++level;
		OCR1A = get_OCR_value_level(level);

	} else {
		lines_until_next_level -= clr_rows;
	}
}

static void load_high_score(void)
{
	high_score = eeprom_read_dword((const uint32_t *)EEPROMADDR);
}

static void save_high_score(void)
{
	if (score > high_score) {
		eeprom_update_dword((uint32_t *)EEPROMADDR, score);
	}
}



void game_init()
{
	srand(generate_seed());
	// Clear board
	memset(game_board, 0, BOARD_ROWS * BOARD_COLS * sizeof(uint8_t));
	level = 0;
	score = 0;
	lines_until_next_level = get_lines_until_level(level);
	state = GAME_START_SCREEN;
	load_high_score();
	spawn_new_tetro();
	timer1_init();
}

/*****************************************************************************/
/*                                State loops                                */
/*****************************************************************************/
static uint8_t any_button_pressed(void)
{
	return up_button_pressed || down_button_pressed ||
		left_button_pressed || right_button_pressed ||
		ccw_button_pressed || cw_button_pressed;
}

static void set_all_buttons_not_pressed(void)
{
	up_button_pressed = 0;
	down_button_pressed = 0;
	left_button_pressed = 0;
	right_button_pressed = 0;
	ccw_button_pressed = 0;
	cw_button_pressed = 0;
}

static void start_screen_loop(void)
{
	if (any_button_pressed()) {
		state = GAME_RUNNING;
		set_all_buttons_not_pressed();
	}
}

static void running_loop(void)
{
	if (down_button_pressed || down_timer_expired || down_button_hold) {
		if (!tetro_move_down()) {
			int clr_rows;

			land_tetro();
			clr_rows = clear_filled_rows();
			score_add_clear_rows(clr_rows);
			level_up_if_needed(clr_rows);

			spawn_new_tetro();
			if (test_collision(curr_tetro.row, curr_tetro.col,
				curr_tetro.rot)) {

				save_high_score();
				state = GAME_END_SCREEN;
			}
		} else if (down_button_hold){
			score_add_drop_points();
		}
		down_button_pressed = 0;
		down_timer_expired = 0;
		down_button_hold = 0;
	}

	if (left_button_pressed || left_button_hold) {
		tetro_move_left();
		left_button_pressed = 0;
		left_button_hold = 0;
	}

	if (right_button_pressed || right_button_hold) {
		tetro_move_right();
		right_button_pressed = 0;
		right_button_hold = 0;
	}

	if (ccw_button_pressed || ccw_button_hold) {
		tetro_rotate_ccw();
		ccw_button_pressed = 0;
		ccw_button_hold = 0;
	}

	if (cw_button_pressed || ccw_button_hold) {
		tetro_rotate_cw();
		cw_button_pressed = 0;
		ccw_button_hold = 0;
	}

	if (up_button_pressed) {
		state = GAME_PAUSED;
		up_button_pressed = 0;
	}
}

static void paused_loop(void)
{
	if (any_button_pressed()) {
		state = GAME_RUNNING;
		set_all_buttons_not_pressed();
	}
}

static void end_loop(void)
{
	if (any_button_pressed()) {
		game_init();
		state = GAME_RUNNING;
		set_all_buttons_not_pressed();
	}
}

void game_loop(void)
{
	switch (state) {
	case GAME_START_SCREEN:
		start_screen_loop();
		break;

	case GAME_RUNNING:
		running_loop();
		break;

	case GAME_PAUSED:
		paused_loop();
		break;

	case GAME_END_SCREEN:
		end_loop();
		break;
	}

}

/*****************************************************************************/
/*                           Asynchronous Functions                          */
/*****************************************************************************/
void game_down_button_pressed(void)
{
	down_button_pressed = 1;
}
void game_down_button_hold(void)
{
	down_button_hold = 1;
}

void game_left_button_pressed(void)
{
	left_button_pressed = 1;
}
void game_left_button_hold(void)
{
	left_button_hold = 1;
}

void game_right_button_pressed(void)
{
	right_button_pressed = 1;
}
void game_right_button_hold(void)
{
	right_button_hold = 1;
}

void game_cw_button_pressed(void)
{
	cw_button_pressed = 1;
}
void game_cw_button_hold(void)
{
	cw_button_hold = 1;
}

void game_ccw_button_pressed(void)
{
	ccw_button_pressed = 1;
}
void game_ccw_button_hold(void)
{
	ccw_button_hold = 1;
}

void game_up_button_pressed(void)
{
	up_button_pressed = 1;
}
void game_up_button_hold(void)
{
	up_button_hold = 1;
}

/*****************************************************************************/
/*                                  Getters                                  */
/*****************************************************************************/
uint8_t *game_get_board(void)
{
	return game_board;
}

struct tetro *game_get_tetro(void)
{
	return &curr_tetro;
}

uint8_t game_get_level(void)
{
	return level;
}

uint32_t game_get_score(void)
{
	return score;
}

uint32_t game_get_high_score(void)
{
	return high_score;
}

enum game_state game_get_state(void)
{
	return state;
}
