#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>

#include "game.h"
#include "renderer.h"

#define PRESS_COUNT		4
#define DAS_AFTER_COUNT		70
#define DAS_INIT_COUNT		200

struct button_data {
	int pin;
	// variable to tell if button is pressed and debounced
	uint8_t is_down;
	// variable to tell if button is hold
	uint8_t is_held;
	// counter for number of equal states
	uint8_t count;
	// counter for number of equal states after button is pressed and
	// debounced
	uint16_t hold_count;
	// keeps track of current debounced state
	uint8_t state;
	// function to signal the game when button is pressed
	void (*pressed_function) (void);
	// function to signal the game when button is held
	void (*held_function) (void);
};

static struct button_data up_button, left_button, right_button, down_button;
static struct button_data ccw_button, cw_button;

/**
 * Check button state and set the is_down variable if a debounced
 * button down press is detected.
 * @param button pointer to button to be debounced
 */
static void debounce_button(struct button_data *button)
{
	// Check if button is high or low for the moment
	uint8_t curr_state = ((PINB & (1 << (*button).pin)) == 0);

	if (curr_state != (*button).state) {
		// Button state is about to be changed, increase counter
		++(*button).count;
		if ((*button).count >= PRESS_COUNT) {
			// The button have not bounced for four checks,
			// change state
			(*button).state = curr_state;
			// If the button was pressed (not released),
			// tell main so
			if (curr_state != 0) {
				(*button).is_down = 1;
				(*button).is_held = 1;
			} else {
				(*button).is_held = 0;
			}
			(*button).count = 0;
			(*button).hold_count = 0;
		}
	} else {
		// Reset counter
		(*button).count = 0;
		(*button).hold_count++;
	}
}

/**
 * Register button press action and signal game.
 * @param button pointer to button to be debounced
 */
static void handle_button_press(struct button_data *button)
{
	debounce_button(button);
	if ((*button).is_down) {
		(*button).is_down = 0;
		(*button).pressed_function();
	}
}

/**
 * Register button press and hold actions and signal game.
 * @param button pointer to button to be debounced
 */
static void handle_button_hold(struct button_data *button)
{
	debounce_button(button);
	if ((*button).is_down) {
		(*button).is_down = 0;
		(*button).pressed_function();
	}
	// DAS = Delayed Auto Shift, more info:
	// https://tetris.fandom.com/wiki/DAS
	if ((*button).is_held &&
		((*button).hold_count > DAS_INIT_COUNT &&
			(*button).hold_count % DAS_AFTER_COUNT == 0)) {
		(*button).held_function();
	}
}

/**
 * Interrupt every 10 ms.
 */
ISR(TIMER0_COMPA_vect)
{
	handle_button_press(&up_button);
	handle_button_hold(&left_button);
	handle_button_hold(&down_button);
	handle_button_hold(&right_button);
	handle_button_press(&ccw_button);
	handle_button_press(&cw_button);
}

static void init_leds(void)
{
	// Red LED
	DDRD |= (1 << PD7);	// set port as output
	PORTD |= (1 << PD7);	// turn on led
	// Blue Screen LEDs
	DDRC |= (1 << PC7);
	PORTC |= (1 << PC7);
}

static void init_buttons(void)
{
	// Up
	DDRB &= ~(1 << PB4);	// set port as input
	PORTB |= (1 << PB4);	// activate pull-up resistor
	up_button.pin = PB4;
	up_button.pressed_function = game_up_button_pressed;
	up_button.held_function = game_up_button_hold;
	// Left
	DDRB &= ~(1 << PB6);
	PORTB |= (1 << PB6);
	left_button.pin = PB6;
	left_button.pressed_function = game_left_button_pressed;
	left_button.held_function = game_left_button_hold;
	// Down
	DDRB &= ~(1 << PB5);
	PORTB |= (1 << PB5);
	down_button.pin = PB5;
	down_button.pressed_function = game_down_button_pressed;
	down_button.held_function = game_down_button_hold;
	// Right
	DDRB &= ~(1 << PB3);
	PORTB |= (1 << PB3);
	right_button.pin = PB3;
	right_button.pressed_function = game_right_button_pressed;
	right_button.held_function = game_right_button_hold;
	// CCW
	DDRB &= ~(1 << PB1);
	PORTB |= (1 << PB1);
	ccw_button.pin = PB1;
	ccw_button.pressed_function = game_ccw_button_pressed;
	ccw_button.held_function = game_ccw_button_hold;
	// CW
	DDRB &= ~(1 << PB0);
	PORTB |= (1 << PB0);
	cw_button.pin = PB0;
	cw_button.pressed_function = game_cw_button_pressed;
	cw_button.held_function = game_cw_button_hold;
}

static void init_timer0(void)
{
	// 10 ms limit
	OCR0A = 25;
	// activate interruption when timer equals OCR0A
	TIMSK0 = (1 << OCIE0A);
	// set timer in CTC mode
	TCCR0B = (3 << CS00) | (1 << WGM02);
}

static void init(void)
{
	sei();
	init_leds();
	init_buttons();
	init_timer0();
	game_init();
	renderer_init();
}

static void loop(void)
{
	for(;;) {
		game_loop();
		render_scene();
	}
}

int main(void)
{
	init();
	loop();

	return 0;
}
